# Copyright (C) 2014 Canonical Ltd.
# Author: Michael Vogt <michael.vogt@ubuntu.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Acquire of Click packages."""

from __future__ import print_function

__metaclass__ = type
__all__ = [
    'ClickAcquire',
    'ClickAcquireError',
    'ClickAcquireMethod',
    'ClickAcquireMethodPycurl',
    'ClickAcquireMethodUbuntuDownloadManager',
    'ClickAcquireStatus',
    'ClickAcquireStatusMeter',
    'ClickAcquireStatusText',
    ]

import os
import math
import select
import subprocess
import sys
from textwrap import dedent

from progressbar import ProgressBar, Bar, Widget

import apt_pkg
from debian.deb822 import Deb822
import dbus
from gi.repository import GLib
import pycurl
from six.moves.urllib.parse import urlparse

from click.paths import acquire_methods_dir
from click.repository import get_repository
from click.oauth import get_oauth_headers_for_uri

class ClickAcquireError(Exception):
    """Error during acquire"""
    pass


def extract_one_message(lines):
    msg = []
    while True:
        try:
            line = lines.pop(0)
        except IndexError:
            break
        if os.environ.get("CLICK_DEBUG_ACQUIRE", ""):
            sys.stderr.write("[%s] raw_line: '%s'\n" % (
                os.getpid(), line.replace("\n", "\\n")))
        msg.append(line)
        if len(msg) > 0 and line == "":
            # we are done, collect all remaining "\n" and stop
            while len(lines) > 0 and lines[0] == "":
                lines.pop(0)
            break
    if len(msg) < 2:
        return -1 , ""
    if os.environ.get("CLICK_DEBUG_ACQUIRE", ""):
        sys.stderr.write("[%s] msg: '%s'\n" % (os.getpid(), str(msg)))
    number = int(msg[0].split()[0])
    return number, "\n".join(msg)


def _read_messages(input_file, timeout=None):
    infd = input_file.fileno()
    msgs = []
    rl, wl, xl = select.select([infd], [], [], timeout)
    if rl:
        # FIXME: 16k message limit is arbitrary
        buf = os.read(infd, 16*1024).decode("utf-8")
        if os.environ.get("CLICK_DEBUG_ACQUIRE", ""):
            sys.stderr.write("[%s] read buf: '%s'\n" % (
                os.getpid(), buf))
        lines = buf.split("\n")
        while True:
            number, msg = extract_one_message(lines)
            if number < 0:
                break
            msgs.append( (number, msg) )
    return msgs


class ClickAcquireStatus:
    """Base class for the status reporting """

    def __init__(self):
        self.fetched_bytes = 0
        self.total_bytes = 1.0
        self.uri = ""

    def pulse(self):
        pass

    def done(self):
        pass


class ClickAcquireStatusText(ClickAcquireStatus):
    """Text based progress reporting for the acquire progress"""

    def pulse(self):
        sys.stdout.write("\r")
        sys.stdout.write("[%3.2f %%] Fetching %s" % (
            (self.fetched_bytes/self.total_bytes)*100.0,
            os.path.basename(self.uri)))
        sys.stdout.flush()

    def done(self):
        self.pulse()
        sys.stdout.write("\n")
        sys.stdout.write("Done fetching %s\n" % self.uri)


class FileSize(Widget):
    def update(self, pbar):
        value = pbar.maxval                         # or .currval?
        if value < 1000:
            return '{:3d} B '.format(int(value))
        scale = int(math.log(value, 1000))          # or binary? e.g. MiB
        try:
            suffix = ' kMGTPEZY'[scale]
        except IndexError:
            # Should never happen, but still.
            suffix = '!'
        scaled_value = value // 1000 ** scale
        return '{:3d} {:1s}B '.format(scaled_value, suffix)


class OK(Widget):
    def update(self, pbar):
        if pbar.currval >= pbar.maxval:
            return 'OK'
        else:
            return '  '

SPACER = '    '

class ClickAcquireStatusMeter(ClickAcquireStatus):
    """Progress meter reporting for the acquire progress"""

    # We have to defer creating the actual progress bar until its first use
    # because the constructor doesn't get called with essential information
    # (e.g. the uri and the total bytes).  These get set after the fact but
    # before the first call to .pulse(), so initialize the object there.
    _progress = None

    def pulse(self):
        if self.total_bytes <= 1:
            return
        if self._progress is None:
            # Massage the uri.  Use the part to the right of the .click
            # suffix, minus any version number.
            base = os.path.basename(self.uri)
            if base.startswith('com.ubuntu.snappy.'):
                base = base[18:]
            snap, _, extra = base.partition('_')
            self._progress = ProgressBar(widgets=[
                snap,
                SPACER,
                # XXX in the examples, it's unclear whether the size field
                # should be the total size or the fetched size.  Also, decimal
                # (MB) or binary (MiB)?  For now, we go with total size in
                # decimal.
                FileSize(),
                SPACER,
                Bar(marker='=', left='[', right=']'),
                SPACER,
                OK(),
                SPACER,
                ],
                maxval=self.total_bytes)
            self._progress.start()
        self._progress.update(self.fetched_bytes)

    def done(self):
        self._progress.finish()


# similar to Acquire/AcquireWorker
class ClickAcquire:
    """Acquire from remote locations"""

    # default status reporting timeout
    TIMEOUT = 0.05

    # acquire method status codecs
    M_CAPABILITIES = 100
    M_STATUS = 102
    M_REDIRECT = 103
    URI_START = 200
    URI_SUCCESS = 201
    URI_FAILURE = 400

    def __init__(self,  status=ClickAcquireStatus()):
        self._fetch_queue = []
        self._status = status

    def _uri_acquire(self, pipe, uri, destfile):
        cmd = dedent("""\
        600 URI Acquire
        URI: {uri}
        Filename: {destfile}

        """).format(uri=uri, destfile=destfile)
        pipe.write(cmd)
        pipe.flush()

    def _redirect(self, pipe, new_uri, destfile):
        self._uri_acquire(pipe, new_uri, destfile)

    def _run_acquire_method(self, uri, destfile):
        parsed_uri = urlparse(uri)
        cmd = os.path.join(acquire_methods_dir, parsed_uri.scheme)
        p = subprocess.Popen([cmd],
            stdout=subprocess.PIPE, stdin=subprocess.PIPE,
            universal_newlines=True)
        while True:
            for number, raw_message in _read_messages(p.stdout, self.TIMEOUT):
                message = Deb822(raw_message)
                if number == self.M_CAPABILITIES:
                    self._uri_acquire(p.stdin, uri, destfile)
                elif number == self.M_STATUS:
                    pass
                elif number == self.M_REDIRECT:
                    self._redirect(p.stdin, message.get("New-URI"), destfile)
                elif number == self.URI_START:
                    self._status.uri = message.get("URI", 0)
                    self._status.total_bytes = int(message.get("Size", 0))
                elif number == self.URI_SUCCESS:
                    self._status.fetched_bytes = int(message.get("Size"))
                    self._status.done()
                    p.stdout.close()
                    p.stdin.close()
                    return True
                elif number == self.URI_FAILURE:
                    p.stdout.close()
                    p.stdin.close()
                    raise ClickAcquireError("Uri failure for %s: %s" % (
                        message.get("uri"), message.get("Message")))
            # update progress
            if os.path.exists(destfile):
                self._status.fetched_bytes = os.path.getsize(destfile)
                self._status.pulse()
        return False

    def fetch(self, uri, destfile):
        self._run_acquire_method(uri, destfile)


# similar to the apt AcquireMethod
class ClickAcquireMethod:

    M_CONFIGURATION = 601
    M_FETCH = 600

    VERSION = 0.1

    def __init__(self):
        s = dedent("""\
        100 Capabilities
        Version: {version}
        Single-Instance: true

        """).format(version=self.VERSION)
        sys.stdout.write(s)
        sys.stdout.flush()

    def run(self):
        while True:
            msgs = _read_messages(sys.stdin)
            if not msgs:
                break
            for number, raw_message in msgs:
                message = Deb822(raw_message)
                if number == self.M_CONFIGURATION:
                    pass
                elif number == self.M_FETCH:
                    self.fetch(message.get("URI"), message.get("FileName"))

    def uri_start(self, uri, filename, size):
        # note that apt itself does not use "Filename" here because it
        # will set a filename and expects the method to use it, however
        # this will not work with UbuntuDownloadManager as it downloads
        # to its own location
        sys.stdout.write(dedent("""\
        200 URI Start
        URI: {uri}
        Filename: {filename}
        Size: {size}

        """).format(uri=uri, filename=filename, size=size))
        sys.stdout.flush()

    def uri_done(self, uri, filename):
        # bug in python-apt
        hashes = apt_pkg.Hashes(filename.encode("utf-8"))
        sys.stdout.write(dedent("""\
        201 URI Done
        URI: {uri}
        Filename: {filename}
        Size: {size}
        Sha256-Hash: {sha256hash}

        """).format(uri=uri, filename=filename, size=os.path.getsize(filename),
                    sha256hash=hashes.sha256))
        sys.stdout.flush()

    def fail(self, uri, err="unknown error"):
        sys.stdout.write(dedent("""\
        400 URI Failure
        URI: {uri}
        Message: {err}

        """).format(uri=uri, err=err))
        sys.stdout.flush()

    def fetch(self, uri, destfile):
        pass


class ClickAcquireMethodPycurl(ClickAcquireMethod):

    def _write_to_file_callback(self, data):
        self._destfile_fp.write(data)
        return len(data)

    def _progress_info(self, dltotal, dlnow, ultotal, ulnow):
        if dltotal > 0 and not self._uri_start_reported:
            self.uri_start(
                self.uri, self._destfile_fp.name, int(dltotal))
            self._uri_start_reported = True

    def fetch(self, uri, destfile):
        self._uri_start_reported = False
        self.uri = uri
        self._destfile_fp = open(destfile, "wb")
        curl = pycurl.Curl()
        curl.setopt(pycurl.URL, uri)
        curl.setopt(pycurl.WRITEFUNCTION, self._write_to_file_callback)
        curl.setopt(pycurl.NOPROGRESS, 0)
        curl.setopt(pycurl.PROGRESSFUNCTION, self._progress_info)
        curl.setopt(pycurl.FOLLOWLOCATION, 1)
        curl.setopt(pycurl.MAXREDIRS, 5)
        curl.setopt(pycurl.FAILONERROR, 1)
        # timeout 120s for conenction
        curl.setopt(pycurl.CONNECTTIMEOUT, 120)
        # timeout if the speed is 120s below 10 bytes/sec
        curl.setopt(pycurl.LOW_SPEED_LIMIT, 10);
        curl.setopt(pycurl.LOW_SPEED_TIME, 120);
        # ssl: no need to set any option here,
        #       SSL_VERIFYPEER=1, SSL_VERIFYHOST=2,
        #       CAINFO=/etc/ssl/certs/ca-certificates.crt
        #      by default in libcurl3 these days

        # FIXME: duplicated code with network.py:http_request
        # add auth header
        repo = get_repository()
        if repo.credentials:
            headers = get_oauth_headers_for_uri(uri, repo.credentials)
            header_list = []
            for k, v in headers.items():
                header_list.append("%s: %s" % (k, v))
            curl.setopt(pycurl.HTTPHEADER, header_list)
        # do it
        try:
            curl.perform()
            self._destfile_fp.close()
            self.uri_done(uri, destfile)
        except pycurl.error as e:
            self.fail(self.uri, e.args)
            self._destfile_fp.close()


class ClickAcquireMethodUbuntuDownloadManager(ClickAcquireMethod):

    MANAGER_PATH = '/'
    MANAGER_IFACE = 'com.canonical.applications.DownloadManager'
    DOWNLOAD_IFACE = 'com.canonical.applications.Download'

    def __init__(self):
        super(ClickAcquireMethodUbuntuDownloadManager, self).__init__()
        self.bus = dbus.SessionBus()
        self.loop = GLib.MainLoop()

    def _created_callback(self, dbus_path):
        pass

    def _finished_callback(self, path, loop):
        self.down_path = path
        loop.quit()

    def _progress_callback(self, total, progress):
        #print('Progress is %s/%s' % (progress, total))
        if not self._started:
            # FIXME: we need the tmpfile path from udm
            tmpfile = "meep"
            self.uri_start(self.uri, tmpfile, total)
            self._started = True

    def fetch(self, uri, destfile):
        self.uri = uri
        self._started = False
        manager = self.bus.get_object(
            'com.canonical.applications.Downloader', self.MANAGER_PATH)
        manager_dev_iface = dbus.Interface(
            manager, dbus_interface=self.MANAGER_IFACE)
        manager_dev_iface.connect_to_signal(
            'downloadCreated', self._created_callback)
        down_path = manager_dev_iface.createDownload(
            (uri, "", "", dbus.Dictionary({}, signature="sv"),
             dbus.Dictionary({}, signature="ss")))
        download1 = self.bus.get_object('com.canonical.applications.Downloader',
            down_path)
        download_dev_iface1 = dbus.Interface(
            download1, dbus_interface=self.DOWNLOAD_IFACE)
        download_dev_iface1.connect_to_signal(
            'progress', self._progress_callback)
        download_dev_iface1.connect_to_signal(
            'finished',
            lambda path: self._finished_callback(path, self.loop))
        download_dev_iface1.start()
        self.loop.run()
        # FIXME error handling, i.e. send self.fail()
        self.uri_done(uri, self.down_path)


if __name__ == "__main__":
    log = ClickAcquireStatusText()
    acq = ClickAcquire(log)
    uri = sys.argv[1]
    if not acq.fetch(uri, os.path.basename(uri)):
        sys.exit(1)
    sys.exit(0)
