# Copyright (C) 2013 Canonical Ltd.
# Author: Colin Watson <cjwatson@ubuntu.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Install a Click package (low-level; consider pkcon instead)."""

from __future__ import print_function

from optparse import OptionParser
import os
import sys
import tempfile
from textwrap import dedent

from six.moves.urllib.parse import urlparse

from gi.repository import Click

from click.acquire import (
    ClickAcquire,
    ClickAcquireStatusMeter,
    ClickAcquireStatusText,
)
from click.install import ClickInstaller, ClickInstallerError
from click.repository import (
    get_repository,
)

def download_click(package_uri, progress_class=None):
    if progress_class is None:
        # For backward compatibility.
        progress_class = ClickAcquireStatusText
    t = tempfile.NamedTemporaryFile()
    package_path = t.name
    log = progress_class()
    acq = ClickAcquire(log)
    acq.fetch(package_uri, package_path)
    # return the tempfile so that it does not get deleted when we
    # go out of scope here
    return t


def run(argv):
    parser = OptionParser(dedent("""\
        %prog install [options] PACKAGE-FILE

        This is a low-level tool; to install a package as an ordinary user
        you should generally use "pkcon install-local PACKAGE-FILE"
        instead."""))
    parser.add_option(
        "--root", metavar="PATH", help="install packages underneath PATH")
    parser.add_option(
        "--force-missing-framework", action="store_true", default=False,
        help="install despite missing system framework")
    parser.add_option(
        "--user", metavar="USER", help="register package for USER")
    parser.add_option(
        "--all-users", default=False, action="store_true",
        help="register package for all users")
    parser.add_option(
        "--allow-unauthenticated", default=False, action="store_true",
        help="allow installing packages with no signatures")
    parser.add_option(
        '--meter', default=False, action='store_true',
        help='display progress with a text meter')
    parser.add_option(
        "--verbose", default=False, action="store_true",
        help="be more verbose on install")
    options, args = parser.parse_args(argv)
    if len(args) < 1:
        parser.error("need package file name")
    progress_class = (ClickAcquireStatusMeter
                      if options.meter else ClickAcquireStatusText)
    db = Click.DB()
    db.read(db_dir=None)
    if options.root is not None:
        db.add(options.root)
    package_uri = args[0]
    installer = ClickInstaller(
        db=db, force_missing_framework=options.force_missing_framework,
        allow_unauthenticated=options.allow_unauthenticated)
    parsed_uri = urlparse(package_uri)
    if parsed_uri.scheme != "":
        t = download_click(package_uri, progress_class)
        package_path = t.name
    elif os.path.exists(package_uri):
        package_path = package_uri
    else:
        repo = get_repository()
        details = repo.details(package_uri)
        if repo.credentials:
            t = download_click(details["download_url"], progress_class)
        else:
            t = download_click(details["anon_download_url"], progress_class)
        package_path = t.name
    try:
        installer.install(
            package_path, user=options.user, all_users=options.all_users,
            quiet=not options.verbose)
    except ClickInstallerError as e:
        print("Cannot install %s: %s" % (package_path, e), file=sys.stderr)
        return 1
    return 0
