# Copyright (C) 2014 Canonical Ltd.
# Author: Michael Vogt

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Update Click packages"""

from __future__ import print_function

from optparse import OptionParser
import sys
import tempfile
from textwrap import dedent

from prettytable import PrettyTable, PLAIN_COLUMNS

from gi.repository import Click
from click.acquire import (
    ClickAcquire,
    ClickAcquireStatusMeter,
    ClickAcquireStatusText,
)
from click.install import ClickInstaller

from click.repository import (
    get_repository,
    ClickRepositoryError,
)


def run(argv):
    parser = OptionParser(dedent("""\
        %prog update

        """))
    parser.add_option(
        "--list", default=False, action="store_true",
        help="Just list the update")
    parser.add_option(
        "--machine-readable", default=False, action="store_true",
        help="output in machine readable form")
    parser.add_option(
        "--user", metavar="USER", help="register package for USER")
    parser.add_option(
        "--all-users", default=False, action="store_true",
        help="register package for all users")
    parser.add_option(
        '--meter', default=False, action='store_true',
        help='display progress with a text meter')
    options, args = parser.parse_args(argv)
    repo = get_repository()
    try:
        all_updates = repo.get_upgradable()
    except ClickRepositoryError as e:
        print("Update failed: %s" % e, file=sys.stderr)
        return 1
    if not all_updates:
        return 0

    if len(args) > 0:
        updates = [up for up in all_updates if up[0]["name"] in args]
    else:
        updates = all_updates

    progress_class = (ClickAcquireStatusMeter
                      if options.meter else ClickAcquireStatusText)
    if options.machine_readable:
        for (current, new) in updates:
            print("%s|%s|%s" % (
                new["name"], current["version"], new["version"]))
    elif options.list:
        t = PrettyTable(["Part","Installed", "Available"])
        t.set_style(PLAIN_COLUMNS)
        t.align["Part"] = "l"
        for (current, new) in updates:
            t.add_row((new["name"], current["version"], new["version"]))
        print(t)
    else:
        db = Click.DB()
        db.read(db_dir=None)
        installer = ClickInstaller(db=db)
        for (current, new) in updates:
            if repo.credentials:
                package_uri = new["download_url"]
            else:
                package_uri = new["anon_download_url"]
            print(package_uri)
            # FIXME: duplication from commands/install.py
            t = tempfile.NamedTemporaryFile()
            package_path = t.name
            log = progress_class()
            acq = ClickAcquire(log)
            acq.fetch(package_uri, package_path)
            installer.install(
                package_path, user=options.user, all_users=options.all_users)
        # output what just happend
        t = PrettyTable(["Part","Installed"])
        t.set_style(PLAIN_COLUMNS)
        t.align["Part"] = "l"
        for (current, new) in updates:
            t.add_row((new["name"], new["version"]))
        print(t)

    return 0
