# Copyright (C) 2014 Canonical Ltd.
# Author: Michael Vogt

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Search Click packages."""

from __future__ import print_function

__metaclass__ = type
__all__ = [
    'ClickRepository',
    'ClickRepositoryError',
    ]

import json
import platform
import os

import apt_pkg
from gi.repository import Click
from six.moves.urllib.parse import quote as url_quote

from click.json_helpers import json_array_to_python

import click.install
from click.network import http_request
from click.oauth import get_oauth_headers_for_uri


class ClickLoginTwoFactorRequired(Exception):
    pass


class ClickRepositoryError(Exception):
    pass


def get_repository():
    """Factory to get a click repository for the current distro"""
    distro = platform.dist()[0]
    if distro.startswith('Ubuntu'):
        return ClickRepositoryAppsUbuntuCom()
    raise ClickRepositoryError("No repository for distribution '%s'" % distro)


class ClickRepository:

    def search(self, search_string):
        """Search the repository for the given search string

        Returns a list of click manifest from the repository
        """
        raise NotImplementedError()

    def get_upgradable(self):
        """Get all click packages where a newer version is available

        Returns a list of (old_manifest, new_manifest) pairs
        """
        raise NotImplementedError()

    def details(self, name):
        """Get the details for the given click name

        Returns a click manifest
        """
        raise NotImplementedError()

    def login(self, username, password, otp=None, store_on_disk=True):
        """Log into the given repository with username/password"""
        raise NotImplementedError()

    @property
    def description(self):
        """The description of the repository as a URL"""
        raise NotImplementedError()

    @property
    def credentials(self):
        """The credentials for the repository (or None)"""
        return None


def get_store_headers():
    frameworks = ["%s-%s" % (f.get_base_name(), f.get_base_version())
                  for f in Click.Framework.get_frameworks()]
    headers = {
        "Accept": "application/hal+json",
        "X-Ubuntu-Frameworks": ",".join(frameworks),
        "X-Ubuntu-Architecture": click.install._dpkg_architecture(),
    }
    return headers


class ClickRepositoryAppsUbuntuCom(ClickRepository):
    """Interface to click repository from apps.ubuntu.com"""

    SEARCH_URI = "https://search.apps.ubuntu.com/api/v1/search?q=%s"
    DETAILS_URI = "https://search.apps.ubuntu.com/api/v1/package/%s"
    BULK_APPS_URI = "https://myapps.developer.ubuntu.com/dev/api/"\
                    "click-metadata/"

    TOKEN_FILE = os.path.expanduser("~/.config/click/ubuntu-sso.json")

    def _http_request(self, uri, *args, **kwargs):
        """Send a oauth signed http request to the repo server"""
        if self.credentials:
            token = self.credentials
            headers = kwargs.pop("headers", {})
            headers.update(get_oauth_headers_for_uri(uri, token))
            kwargs["headers"] = headers
        return http_request(uri, *args, **kwargs)

    @property
    def credentials(self):
        if os.path.exists(self.TOKEN_FILE):
            with open(self.TOKEN_FILE) as f:
                return json.load(f)
        return None

    @property
    def description(self):
        return "https://myapps.developer.ubuntu.com/dev/click-apps/"

    def search(self, search_term):
        result = []
        url = self.SEARCH_URI % url_quote(search_term)
        if not self.credentials:
            url += ",allow_unauthenticated:True"
        resp, raw_content = http_request(url, headers=get_store_headers())
        content = raw_content.decode("utf-8")
        if resp.status != 200:
            raise ClickRepositoryError(content)
        data = json.loads(content)
        if not "_embedded" in data:
            return result
        for item in data["_embedded"]["clickindex:package"]:
            result.append(item)
        return result

    def details(self, name):
        resp, raw_content = self._http_request(
            self.DETAILS_URI % url_quote(name), headers=get_store_headers())
        content = raw_content.decode("utf-8")
        if resp.status != 200:
            raise ClickRepositoryError(content)
        return json.loads(content)

    def get_upgradable(self):
        result = []
        db = Click.DB()
        db.read(db_dir=None)
        registry = Click.User.for_user(db, name=None)
        # build the dict of installed apps for the POST
        current_apps = {}
        for app in json_array_to_python(registry.get_manifests()):
            name = app["name"]
            # FIXME: THIS MAKES BABY JESUS CRY
            if not "." in name:
                name = "com.ubuntu.snappy." + name
            current_apps[name] = app
        # if we have nothing installed, do not query the server
        if not current_apps:
            return result
        json_post_data = json.dumps({
            "name": list(current_apps.keys()),
        }).encode("utf-8")
        # query the API
        resp, raw_content = http_request(
            self.BULK_APPS_URI,
            json_post_data,
            headers={"content-type": "application/json"})
        content = raw_content.decode("utf-8")
        if resp.status != 200:
            raise ClickRepositoryError(
                "Failed to get upgradable packages: %s" % resp.content)
        # return a list of (old, new) pairs to the caller
        server_apps = json.loads(content)
        result = []
        for new in server_apps:
            current = current_apps[new["name"]]
            if apt_pkg.version_compare(current["version"], new["version"]) < 0:
                result.append( (current, new) )
        return result

    def login(self, username, password, otp=None, store_on_disk=True):
        token_name = "click"
        token = self._get_oauth_token(username, password, token_name, otp)

        if store_on_disk:
            Click.ensuredir(os.path.dirname(self.TOKEN_FILE))
            with open(self.TOKEN_FILE, "w") as f:
                os.chmod(f.name, 0o600)
                f.write(json.dumps(token))

        return token

    def _get_oauth_token(self, username, password, token_name, otp=None):
        UBUNTUONE_API_BASE = "https://login.ubuntu.com/api/v2"
        UBUNTUONE_OAUTH_API = UBUNTUONE_API_BASE + "/tokens/oauth"

        data = {"email": username,
                "password": password,
                "token_name": token_name,
            }
        if otp:
            data["otp"] = otp
        data = json.dumps(data).encode("utf-8")
        resp, content = http_request(
            UBUNTUONE_OAUTH_API,
            data=data,
            headers={"content-type": "application/json"})
        msg = json.loads(content.decode("utf-8"))
        if resp.status not in (200, 201):
            if msg["code"] == "TWOFACTOR_REQUIRED":
                raise ClickLoginTwoFactorRequired()
            else:
                raise Exception(
                    "get_oauth_token failed with status %s" % resp.status)
        token = msg
        return token
