# Copyright (C) 2014 Canonical Ltd.
# Author: Michael Vogt <michael.vogt@ubuntu.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Unit tests for click.acquire."""

from __future__ import print_function

__metaclass__ = type
__all__ = [
    'TestClickAcquire',
    ]

import multiprocessing
import os.path
import time

from six.moves import (
    SimpleHTTPServer,
    socketserver,
)

from click.acquire import (
    ClickAcquire,
    ClickAcquireError,
    _read_messages,
)

from click.tests.helpers import (
    TestCase,
)

import click.acquire
click.acquire.acquire_methods_dir = os.path.join(
    os.path.dirname(__file__), "..", "..", "acquire")
os.environ["PYTHONPATH"] = os.path.abspath(
    os.path.join(os.path.dirname(__file__), "..", ".."))

# local httpd
LOCALHOST = "localhost"
PORT = 8128


class MyHandler(SimpleHTTPServer.SimpleHTTPRequestHandler):
    def log_message(self, format, *args):
        pass


class MySocketServer(socketserver.TCPServer):
    allow_reuse_address = True


class Httpd(multiprocessing.Process):

    def __init__(self, basedir):
        super(Httpd, self).__init__()
        self.basedir = basedir

    def run(self):
        os.chdir(self.basedir)
        server = MySocketServer((LOCALHOST, PORT), MyHandler)
        while True:
            server.handle_request()
        server.shutdown()

    def stop(self):
        self.terminate()
        self.join()


class TestClickAcquire(TestCase):

    def setUp(self):
        super(TestClickAcquire, self).setUp()
        self.use_temp_dir()
        self.httpd = Httpd(self.temp_dir)
        self.httpd.start()

    def tearDown(self):
        self.httpd.stop()

    def test_acquire_fail(self):
        acq = ClickAcquire()
        destfile = os.path.join(self.temp_dir, "meep")
        with self.assertRaises(ClickAcquireError):
            acq.fetch("http://%s:%s/not-here" % (LOCALHOST, PORT), destfile)

    def test_acquire_good(self):
        acq = ClickAcquire()
        destfile = os.path.join(self.temp_dir, "meep")
        canary_str = "hello"
        with open(os.path.join(self.temp_dir, "i-am-here"), "w") as f:
            f.write(canary_str)
        acq.fetch("http://%s:%s/i-am-here" % (LOCALHOST, PORT), destfile)
        with open(destfile) as f:
            data = f.read()
        self.assertEqual(canary_str, data)


class TestClickAcquireReadMessages(TestCase):

    def test_forked_read_message(self):
        read_end, write_end = os.pipe()
        pid = os.fork()
        if pid == 0:
            os.close(write_end)
            with os.fdopen(read_end) as f:
                number, msg = _read_messages(f)[0]
                self.assertEqual(msg, "102 Status\nFoo: Bar\n")
                self.assertEqual(number, 102)
            os._exit(0)
        os.close(read_end)
        os.write(write_end, "102 Status\nFoo: Bar\n\n".encode("utf-8"))
        os.waitpid(pid, 0)
        os.close(write_end)

    def test_forked_read_multiple_messages(self):
        read_end, write_end = os.pipe()
        pid = os.fork()
        if pid == 0:
            os.close(write_end)
            with os.fdopen(read_end) as f:
                msgs = _read_messages(f)
                self.assertEqual(len(msgs), 2)
                self.assertEqual(msgs[0][0], 100)
                self.assertEqual(msgs[1][0], 200)
            os._exit(0)
        os.close(read_end)
        os.write(write_end, "100 Status\n\n200 meep\n\n".encode("utf-8"))
        os.waitpid(pid, 0)
        os.close(write_end)

    def test_forked_read_message_broken_pipe(self):
        read_end, write_end = os.pipe()
        pid = os.fork()
        if pid == 0:
            os.close(write_end)
            with os.fdopen(read_end) as f:
                self.assertEqual(_read_messages(f), [])
            os._exit(0)
        os.close(read_end)
        os.write(write_end, "102 close-before-msg".encode("utf-8"))
        os.close(write_end)
        os.waitpid(pid, 0)

    def test_forked_read_message_with_timeout(self):
        read_end, write_end = os.pipe()
        pid = os.fork()
        if pid == 0:
            os.close(write_end)
            with os.fdopen(read_end) as f:
                self.assertEqual(_read_messages(f, 0.1), [])
                msgs = _read_messages(f)
                self.assertEqual(len(msgs), 2)
                self.assertEqual(msgs[0][1], "100 aaa\n")
                self.assertEqual(msgs[1][1], "200 bbb\n")
            os._exit(0)
        os.close(read_end)
        # simulate slow method
        time.sleep(0.2)
        os.write(write_end, "100 aaa\n\n".encode("utf-8"))
        os.write(write_end, "200 bbb\n\n".encode("utf-8"))
        os.waitpid(pid, 0)
