# Copyright (C) 2013 Canonical Ltd.
# Author: Colin Watson <cjwatson@ubuntu.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Unit tests for click.build."""

from __future__ import print_function

__metaclass__ = type
__all__ = [
    'TestClickBuildYaml',
    'TestClickBuilder',
    'TestClickFrameworkValidation',
    'TestClickSourceBuilder',
    ]


import io
import json
import os
import stat
import subprocess
import tarfile
from textwrap import dedent

import yaml

from click.build import ClickBuildError, ClickBuilder, ClickSourceBuilder
from click.preinst import static_preinst
from click.tests.helpers import (
    disable_logging,
    mkfile,
    TestCase,
    touch,
)


# BAW 2013-04-15: Some tests require umask 022.  Use this decorator to
# temporarily tweak the process's umask.  The test -- or system -- should
# probably be made more robust instead.
def umask(force_umask):
    def decorator(func):
        def wrapper(*args, **kws):
            old_umask = os.umask(force_umask)
            try:
                return func(*args, **kws)
            finally:
                os.umask(old_umask)
        return wrapper
    return decorator


def extract_field(path, name):
    return subprocess.check_output(
        ["dpkg-deb", "-f", path, name],
        universal_newlines=True).rstrip("\n")


class TestClickBuilderBaseMixin:
    def test_read_manifest(self):
        self.use_temp_dir()
        manifest_path = os.path.join(self.temp_dir, "manifest.json")
        with mkfile(manifest_path) as manifest:
            print(dedent("""\
                {
                    "name": "com.example.test",
                    "version": "1.0",
                    "maintainer": "Foo Bar <foo@example.org>",
                    "title": "test title",
                    "framework": "ubuntu-sdk-13.10"
                }"""), file=manifest)
        self.builder.read_manifest(manifest_path)
        self.assertEqual("com.example.test", self.builder.name)
        self.assertEqual("1.0", self.builder.version)
        self.assertEqual("Foo Bar <foo@example.org>", self.builder.maintainer)
        self.assertEqual("test title", self.builder.title)
        self.assertEqual("all", self.builder.architecture)

    def test_add_file(self):
        self.builder.add_file("/nonexistent", "target")
        self.assertEqual({"/nonexistent": "target"}, self.builder.file_map)

    def test_epochless_version(self):
        self.use_temp_dir()
        manifest_path = os.path.join(self.temp_dir, "manifest.json")
        for version, epochless_version in (
            ("1.0", "1.0"),
            ("1:1.2.3", "1.2.3"),
        ):
            with mkfile(manifest_path) as manifest:
                print(dedent("""\
                    {
                        "name": "com.example.test",
                        "version": "%s",
                        "maintainer": "Foo Bar <foo@example.org>",
                        "title": "test title",
                        "framework": "ubuntu-sdk-13.10"
                    }""") % version, file=manifest)
            self.builder.read_manifest(manifest_path)
            self.assertEqual(epochless_version, self.builder.epochless_version)

    def test_manifest_syntax_error(self):
        self.use_temp_dir()
        manifest_path = os.path.join(self.temp_dir, "manifest.json")
        with mkfile(manifest_path) as manifest:
            # The comma after the "name" entry is intentionally missing.
            print(dedent("""\
                {
                    "name": "com.example.test"
                    "version": "1.0"
                }"""), file=manifest)
        self.assertRaises(
            ClickBuildError, self.builder.read_manifest, manifest_path)


class TestClickBuilder(TestCase, TestClickBuilderBaseMixin):
    def setUp(self):
        super(TestClickBuilder, self).setUp()
        self.builder = ClickBuilder()

    @disable_logging
    @umask(0o22)
    def test_build(self):
        self.use_temp_dir()
        scratch = os.path.join(self.temp_dir, "scratch")
        with mkfile(os.path.join(scratch, "bin", "foo")) as f:
            f.write("test /bin/foo\n")
        os.symlink("foo", os.path.join(scratch, "bin", "bar"))
        touch(os.path.join(scratch, ".git", "config"))
        with mkfile(os.path.join(scratch, "toplevel")) as f:
            f.write("test /toplevel\n")
        with mkfile(os.path.join(scratch, "manifest.json")) as f:
            json.dump({
                "name": "com.example.test",
                "version": "1.0",
                "maintainer": "Foo Bar <foo@example.org>",
                "title": "test title",
                "architecture": "all",
                "framework": "ubuntu-sdk-13.10",
            }, f)
            # build() overrides this back to 0o644
            os.fchmod(f.fileno(), 0o600)
        self.builder.add_file(scratch, "/")
        path = os.path.join(self.temp_dir, "com.example.test_1.0_all.click")
        self.assertEqual(path, self.builder.build(self.temp_dir))
        self.assertTrue(os.path.exists(path))
        for key, value in (
            ("Package", "com.example.test"),
            ("Version", "1.0"),
            ("Click-Version", "0.4"),
            ("Architecture", "all"),
            ("Maintainer", "Foo Bar <foo@example.org>"),
            ("Description", "test title"),
        ):
            self.assertEqual(value, extract_field(path, key))
        self.assertNotEqual(
            "", extract_field(path, "Installed-Size"))
        control_path = os.path.join(self.temp_dir, "control")
        subprocess.check_call(["dpkg-deb", "-e", path, control_path])
        manifest_path = os.path.join(control_path, "manifest")
        self.assertEqual(0o644, stat.S_IMODE(os.stat(manifest_path).st_mode))
        with open(os.path.join(scratch, "manifest.json")) as source, \
                open(manifest_path) as target:
            source_json = json.load(source)
            target_json = json.load(target)
            self.assertNotEqual("", target_json["installed-size"])
            del target_json["installed-size"]
            self.assertEqual(source_json, target_json)
        with open(os.path.join(control_path, "md5sums")) as md5sums:
            self.assertRegex(
                md5sums.read(),
                r"^"
                r"eb774c3ead632b397d6450d1df25e001  bin/bar\n"
                r"eb774c3ead632b397d6450d1df25e001  bin/foo\n"
                r"49327ce6306df8a87522456b14a179e0  toplevel\n"
                r"$")
        with open(os.path.join(control_path, "preinst")) as preinst:
            self.assertEqual(static_preinst, preinst.read())
        contents = subprocess.check_output(
            ["dpkg-deb", "-c", path], universal_newlines=True)
        self.assertRegex(contents, r"^drwxr-xr-x root/root         0 .* \./\n")
        self.assertRegex(
            contents,
            "\nlrwxrwxrwx root/root         0 .* \./bin/bar -> foo\n")
        self.assertRegex(
            contents, "\n-rw-r--r-- root/root        14 .* \./bin/foo\n")
        self.assertRegex(
            contents, "\n-rw-r--r-- root/root        15 .* \./toplevel\n")
        extract_path = os.path.join(self.temp_dir, "extract")
        subprocess.check_call(["dpkg-deb", "-x", path, extract_path])
        for rel_path in (
            os.path.join("bin", "foo"),
            "toplevel",
        ):
            with open(os.path.join(scratch, rel_path)) as source, \
                    open(os.path.join(extract_path, rel_path)) as target:
                self.assertEqual(source.read(), target.read())
        self.assertTrue(
            os.path.islink(os.path.join(extract_path, "bin", "bar")))
        self.assertEqual(
            "foo", os.readlink(os.path.join(extract_path, "bin", "bar")))

    @disable_logging
    def test_build_excludes_dot_click(self):
        self.use_temp_dir()
        scratch = os.path.join(self.temp_dir, "scratch")
        touch(os.path.join(scratch, ".click", "evil-file"))
        with mkfile(os.path.join(scratch, "manifest.json")) as f:
            json.dump({
                "name": "com.example.test",
                "version": "1.0",
                "maintainer": "Foo Bar <foo@example.org>",
                "title": "test title",
                "architecture": "all",
                "framework": "ubuntu-sdk-13.10",
            }, f)
        self.builder.add_file(scratch, "/")
        path = self.builder.build(self.temp_dir)
        extract_path = os.path.join(self.temp_dir, "extract")
        subprocess.check_call(["dpkg-deb", "-x", path, extract_path])
        self.assertEqual([], os.listdir(extract_path))

    @disable_logging
    def test_build_multiple_architectures(self):
        self.use_temp_dir()
        scratch = os.path.join(self.temp_dir, "scratch")
        with mkfile(os.path.join(scratch, "manifest.json")) as f:
            json.dump({
                "name": "com.example.test",
                "version": "1.0",
                "maintainer": "Foo Bar <foo@example.org>",
                "title": "test title",
                "architecture": ["armhf", "i386"],
                "framework": "ubuntu-sdk-13.10",
            }, f)
        self.builder.add_file(scratch, "/")
        path = os.path.join(self.temp_dir, "com.example.test_1.0_multi.click")
        self.assertEqual(path, self.builder.build(self.temp_dir))
        self.assertTrue(os.path.exists(path))
        self.assertEqual("multi", extract_field(path, "Architecture"))
        control_path = os.path.join(self.temp_dir, "control")
        subprocess.check_call(["dpkg-deb", "-e", path, control_path])
        manifest_path = os.path.join(control_path, "manifest")
        with open(os.path.join(scratch, "manifest.json")) as source, \
                open(manifest_path) as target:
            source_json = json.load(source)
            target_json = json.load(target)
            del target_json["installed-size"]
            self.assertEqual(source_json, target_json)

    # FIXME: DRY violation with test_build_multiple_architectures etc
    @disable_logging
    def test_build_multiple_frameworks(self):
        self.use_temp_dir()
        scratch = os.path.join(self.temp_dir, "scratch")
        with mkfile(os.path.join(scratch, "manifest.json")) as f:
            json.dump({
                "name": "com.example.test",
                "version": "1.0",
                "maintainer": "Foo Bar <foo@example.org>",
                "title": "test title",
                "architecture": "all",
                "framework":
                    "ubuntu-sdk-14.04-basic, ubuntu-sdk-14.04-webapps",
            }, f)
        self.builder.add_file(scratch, "/")
        path = self.builder.build(self.temp_dir)
        control_path = os.path.join(self.temp_dir, "control")
        subprocess.check_call(["dpkg-deb", "-e", path, control_path])
        manifest_path = os.path.join(control_path, "manifest")
        with open(os.path.join(scratch, "manifest.json")) as source, \
                open(manifest_path) as target:
            source_json = json.load(source)
            target_json = json.load(target)
            del target_json["installed-size"]
            self.assertEqual(source_json, target_json)


class TestClickFrameworkValidation(TestCase):
    def setUp(self):
        super(TestClickFrameworkValidation, self).setUp()
        self.builder = ClickBuilder()
        for framework_name in ("ubuntu-sdk-13.10",
                               "ubuntu-sdk-14.04-papi",
                               "ubuntu-sdk-14.04-html"):
            self._create_mock_framework_file(framework_name)

    def test_validate_framework_good(self):
        valid_framework_values = (
            "ubuntu-sdk-13.10",
            "ubuntu-sdk-14.04-papi, ubuntu-sdk-14.04-html",
        )
        for framework in valid_framework_values:
            self.builder._validate_framework(framework)

    def test_validate_framework_bad(self):
        invalid_framework_values = (
            #"ubuntu-sdk-13.10, ubuntu-sdk-14.04-papi",
            "ubuntu-sdk-13.10 (>= 13.10)",
            "ubuntu-sdk-13.10 | ubuntu-sdk-14.04",
        )
        for framework in invalid_framework_values:
            with self.assertRaises(ClickBuildError):
                self.builder._validate_framework(framework)


class TestClickSourceBuilder(TestCase, TestClickBuilderBaseMixin):
    def setUp(self):
        super(TestClickSourceBuilder, self).setUp()
        self.builder = ClickSourceBuilder()

    @umask(0o22)
    def test_build(self):
        self.use_temp_dir()
        scratch = os.path.join(self.temp_dir, "scratch")
        touch(os.path.join(scratch, "bin", "foo"))
        touch(os.path.join(scratch, ".git", "config"))
        with mkfile(os.path.join(scratch, "manifest.json")) as f:
            json.dump({
                "name": "com.example.test",
                "version": "1.0",
                "maintainer": "Foo Bar <foo@example.org>",
                "title": "test title",
                "architecture": "all",
                "framework": "ubuntu-sdk-13.10",
            }, f)
            # build() overrides this back to 0o644
            os.fchmod(f.fileno(), 0o600)
        self.builder.add_file(scratch, "./")
        path = os.path.join(self.temp_dir, "com.example.test_1.0.tar.gz")
        self.assertEqual(path, self.builder.build(self.temp_dir))
        self.assertTrue(os.path.exists(path))
        with tarfile.open(path, mode="r:gz") as tar:
            self.assertCountEqual(
                [".", "./bin", "./bin/foo", "./manifest.json"], tar.getnames())
            self.assertTrue(tar.getmember("./bin/foo").isfile())


SIMPLE_TEMPLATE = """\
name: test-yaml
version: 1.0
maintainer: Foo Bar <foo@example.org>
architecture: all
integration:
  app1:
    hookname: hook-file-name
services:
  - name: app1
    description: some description
    start: start-something
    stop: stop-something
binaries:
  - name: path/to/cli1
"""

NO_STOP_SERVICE_TEMPLATE = """\
name: test-yaml
icon: data/icon.svg
version: 1.0
maintainer: Foo Bar <foo@example.org>
services:
  - name: app1
    description: some description
    start: start-something
"""

MULTI_FRAMEWORK_TEMPLATE = """\
name: test-yaml
icon: data/icon.svg
version: 1.0
maintainer: Foo Bar <foo@example.org>
architecture: all
frameworks: docker,meep
integration:
  app1:
    hookname: hook-file-name
services:
  - name: app1
    description: some description
    start: start-something
    stop: stop-something
binaries:
  - name: path/to/cli1
"""

ALREADY_HAS_UBUNTU_FRAMEWORK_TEMPLATE = """\
name: test-yaml
icon: data/icon.svg
version: 1.0
maintainer: Foo Bar <foo@example.org>
architecture: all
frameworks: ubuntu-rocket-8.04, docker
integration:
  app1:
    hookname: hook-file-name
services:
  - name: app1
    description: some description
    start: start-something
    stop: stop-something
binaries:
  - name: path/to/cli1
"""

APPARMOR_1_TEMPLATE = """\
name: test-yaml
icon: data/icon.svg
version: 1.0
maintainer: Foo Bar <foo@example.org>
architecture: all
integration:
  app1:
    hookname: hook-file-name
    apparmor: path/to/app1.apparmor
services:
  - name: app1
    description: some description
    start: start-something
    stop: stop-something
  - name: app2
    description: some description
    start: start-something
    stop: stop-something
binaries:
  - name: path/to/cli1
"""

APPARMOR_2_TEMPLATE = """\
name: test-yaml
icon: data/icon.svg
version: 1.0
maintainer: Foo Bar <foo@example.org>
architecture: all
integration:
  app1:
    hookname: hook-file-name
    apparmor-profile: path/to/app1.apparmor-profile
services:
  - name: app1
    description: some description
    start: start-something
    stop: stop-something
  - name: app2
    description: some description
    start: start-something
    stop: stop-something
binaries:
  - name: path/to/cli1
"""

APPARMOR_3_TEMPLATE = """\
name: test-yaml
icon: data/icon.svg
version: 1.0
maintainer: Foo Bar <foo@example.org>
architecture: all
integration:
  cli1:
    hookname: hook-file-name
    apparmor: path/to/cli1.apparmor-profile
binaries:
  - name: path/to/cli1
  - name: path/to/cli2
"""

APPARMOR_4_TEMPLATE = """\
name: test-yaml
icon: data/icon.svg
version: 1.0
maintainer: Foo Bar <foo@example.org>
architecture: all
integration:
  cli1:
    hookname: hook-file-name
    apparmor-profile: path/to/cli1.apparmor-profile
binaries:
  - name: path/to/cli1
  - name: path/to/cli2
"""


class TestClickBuildYaml(TestCase):
    def setUp(self):
        super(TestClickBuildYaml, self).setUp()
        self.builder = ClickBuilder()

    def _make_metadata(self, scratch, template=SIMPLE_TEMPLATE, readme=None):
        with mkfile(os.path.join(scratch, "meta", "package.yaml")) as f:
            f.write(dedent(template))
        if readme is None:
            readme = dedent("""test title

            Long description that spawns
            two lines.

            Some more random data.
            """)
        with mkfile(os.path.join(scratch, "meta", "readme.md")) as f:
            f.write(readme)

    def test_build_valid_pkg(self):
        self.use_temp_dir()
        scratch = os.path.join(self.temp_dir, "scratch")
        touch(os.path.join(scratch, "bin", "foo"))
        touch(os.path.join(scratch, "data", "icon.svg"))
        self._make_metadata(scratch)
        self.builder.add_file(scratch, "./")
        # FIXME: make it a .snap
        expected_path = os.path.join(self.temp_dir, "test-yaml_1.0_all.click")
        actual_path = self.builder.build(self.temp_dir, "meta/package.yaml")
        self.assertEqual(expected_path, actual_path)
        self.assertTrue(os.path.exists(actual_path))
        # check the fields
        for key, value in (
            ("Package", "test-yaml"),
            ("Version", "1.0"),
            ("Click-Version", "0.4"),
            ("Architecture", "all"),
            ("Maintainer", "Foo Bar <foo@example.org>"),
            ("Description", "test title"),
        ):
            self.assertEqual(value, extract_field(actual_path, key))
        click_contents = subprocess.check_output(
            ["dpkg-deb", "-c", actual_path], universal_newlines=True)
        for file_path in ("./data/icon.svg", "./meta/package.yaml",
                          "./meta/readme.md", "./bin/foo",
                          "./meta/app1.apparmor", "./meta/cli1.apparmor"):
            self.assertIn(file_path, click_contents)

    def test_read_manifest(self):
        self.use_temp_dir()
        self._make_metadata(self.temp_dir)
        self.builder.read_manifest(self.temp_dir+"/meta/package.yaml")
        self.assertEqual("test-yaml", self.builder.name)
        self.assertEqual("1.0", self.builder.version)
        self.assertEqual("Foo Bar <foo@example.org>", self.builder.maintainer)
        self.assertEqual("test title", self.builder.title)
        self.assertEqual("all", self.builder.architecture)
        self.assertEqual(
            "meta/app1.snappy-systemd",
            self.builder.manifest["hooks"]["app1"]["snappy-systemd"])
        self.assertEqual(
            "hook-file-name",
            self.builder.manifest["hooks"]["app1"]["hookname"])
        self.assertEqual(
            "path/to/cli1",
            self.builder.manifest["hooks"]["cli1"]["bin-path"])
        self.assertEqual(
            "meta/cli1.apparmor",
            self.builder.manifest["hooks"]["cli1"]["apparmor"])
        self.assertEqual(
            "meta/app1.apparmor",
            self.builder.manifest["hooks"]["app1"]["apparmor"])
        # this added automatically by snappy buidl
        self.assertEqual(
            "ubuntu-core-15.04-dev1", self.builder.manifest["framework"])

    def disabled_test_read_manifest_multiple_frameworks(self):
        self.use_temp_dir()
        self._make_metadata(self.temp_dir, MULTI_FRAMEWORK_TEMPLATE)
        self.builder.read_manifest(self.temp_dir+"/meta/package.yaml")
        self.assertEqual(
            "ubuntu-core-15.04-dev1, docker, meep",
            self.builder.manifest["framework"])

    def test_create_systemd_manifest_full(self):
        self.use_temp_dir()
        self._make_metadata(self.temp_dir, SIMPLE_TEMPLATE)
        # this will also generate the systemd
        self.builder.read_manifest(self.temp_dir+"/meta/package.yaml")
        # test it
        with open(self.temp_dir+"/meta/app1.snappy-systemd") as fp:
            systemd = yaml.load(fp)
        self.assertEqual("some description", systemd["description"])
        self.assertEqual("start-something", systemd["start"])
        self.assertEqual("stop-something", systemd["stop"])

    def test_create_systemd_manifest_no_stop(self):
        self.use_temp_dir()
        self._make_metadata(self.temp_dir, NO_STOP_SERVICE_TEMPLATE)
        # this will also generate the systemd
        self.builder.read_manifest(self.temp_dir+"/meta/package.yaml")
        # test it
        with open(self.temp_dir+"/meta/app1.snappy-systemd") as fp:
            systemd = yaml.load(fp)
        self.assertEqual("some description", systemd["description"])
        self.assertEqual("start-something", systemd["start"])

    def test_read_manifest_that_has_ubuntu_framework_already(self):
        self.use_temp_dir()
        self._make_metadata(
            self.temp_dir, ALREADY_HAS_UBUNTU_FRAMEWORK_TEMPLATE)
        self.builder.read_manifest(self.temp_dir+"/meta/package.yaml")
        self.assertEqual(
            "ubuntu-rocket-8.04, docker",
            self.builder.manifest["framework"])

    def test_one_explicit_apparmor(self):
        self.use_temp_dir()
        self._make_metadata(self.temp_dir, APPARMOR_1_TEMPLATE)
        self.builder.read_manifest(self.temp_dir + "/meta/package.yaml")
        self.assertEqual(
            "path/to/app1.apparmor",
            self.builder.manifest["hooks"]["app1"]["apparmor"])
        self.assertEqual(
            "meta/app2.apparmor",
            self.builder.manifest["hooks"]["app2"]["apparmor"])
        self.assertEqual(
            "meta/cli1.apparmor",
            self.builder.manifest["hooks"]["cli1"]["apparmor"])

    def test_one_explicit_apparmor_profile(self):
        self.use_temp_dir()
        self._make_metadata(self.temp_dir, APPARMOR_2_TEMPLATE)
        self.builder.read_manifest(self.temp_dir + "/meta/package.yaml")
        self.assertEqual(
            "path/to/app1.apparmor-profile",
            self.builder.manifest["hooks"]["app1"]["apparmor-profile"])
        self.assertIsNone(
            self.builder.manifest["hooks"]["app1"].get('apparmor'))
        self.assertEqual(
            "meta/app2.apparmor",
            self.builder.manifest["hooks"]["app2"]["apparmor"])
        self.assertEqual(
            "meta/cli1.apparmor",
            self.builder.manifest["hooks"]["cli1"]["apparmor"])

    def test_one_explicit_apparmor_file(self):
        self.use_temp_dir()
        scratch = os.path.join(self.temp_dir, "scratch")
        touch(os.path.join(scratch, "bin", "foo"))
        touch(os.path.join(scratch, "data", "icon.svg"))
        apparmor_path = os.path.join(scratch, 'path', 'to', 'app1.apparmor')
        os.makedirs(os.path.dirname(apparmor_path))
        with io.open(apparmor_path, 'w', encoding='utf-8') as fp:
            fp.write(dedent(u"""\
            {
                "template": "default",
                "policy_groups": [
                    "networking"
                ],
                "policy_vendor": "ubuntu-snappy",
                "policy_version": 99.99
            }
            """))
        self._make_metadata(scratch, APPARMOR_1_TEMPLATE)
        self.builder.add_file(scratch, "./")
        # FIXME: make it a .snap
        expected_path = os.path.join(self.temp_dir, "test-yaml_1.0_all.click")
        actual_path = self.builder.build(self.temp_dir, "meta/package.yaml")
        self.assertEqual(expected_path, actual_path)
        self.assertTrue(os.path.exists(actual_path))
        contents = os.path.join(self.temp_dir, 'contents')
        subprocess.check_output(
            ["dpkg-deb", "-R", actual_path, contents], universal_newlines=True)
        self.assertEqual(sorted(os.listdir(contents)),
                         ['DEBIAN', 'bin', 'data', 'meta', 'path'])
        self.assertEqual(
            sorted(os.listdir(os.path.join(contents, 'path', 'to'))),
            ['app1.apparmor'])
        with io.open(os.path.join(contents, 'path', 'to', 'app1.apparmor'),
                     encoding='utf-8') as fp:
            data = fp.read()
        self.assertEqual(data, dedent("""\
            {
                "template": "default",
                "policy_groups": [
                    "networking"
                ],
                "policy_vendor": "ubuntu-snappy",
                "policy_version": 99.99
            }
            """))
        with io.open(os.path.join(contents, 'meta', 'app2.apparmor'),
                     encoding='utf-8') as fp:
            data = fp.read()
        self.assertEqual(data, dedent("""\
            {
                "template": "default",
                "policy_groups": [
                    "networking"
                ],
                "policy_vendor": "ubuntu-snappy",
                "policy_version": 1.3
            }
            """))

    def test_one_explicit_apparmor_profile_file(self):
        self.use_temp_dir()
        scratch = os.path.join(self.temp_dir, "scratch")
        touch(os.path.join(scratch, "bin", "foo"))
        touch(os.path.join(scratch, "data", "icon.svg"))
        apparmor_path = os.path.join(
            scratch, 'path', 'to', 'app1.apparmor-profile')
        os.makedirs(os.path.dirname(apparmor_path))
        with io.open(apparmor_path, 'w', encoding='utf-8') as fp:
            fp.write(dedent(u"""\
            {
                "template": "default",
                "policy_groups": [
                    "networking"
                ],
                "policy_vendor": "ubuntu-snappy",
                "policy_version": 88.88
            }
            """))
        self._make_metadata(scratch, APPARMOR_2_TEMPLATE)
        self.builder.add_file(scratch, "./")
        # FIXME: make it a .snap
        expected_path = os.path.join(self.temp_dir, "test-yaml_1.0_all.click")
        actual_path = self.builder.build(self.temp_dir, "meta/package.yaml")
        self.assertEqual(expected_path, actual_path)
        self.assertTrue(os.path.exists(actual_path))
        contents = os.path.join(self.temp_dir, 'contents')
        subprocess.check_output(
            ["dpkg-deb", "-R", actual_path, contents], universal_newlines=True)
        self.assertEqual(sorted(os.listdir(contents)),
                         ['DEBIAN', 'bin', 'data', 'meta', 'path'])
        self.assertEqual(
            sorted(os.listdir(os.path.join(contents, 'path', 'to'))),
            ['app1.apparmor-profile'])
        with io.open(
                os.path.join(contents, 'path', 'to', 'app1.apparmor-profile'),
                encoding='utf-8') as fp:
            data = fp.read()
        self.assertEqual(data, dedent("""\
            {
                "template": "default",
                "policy_groups": [
                    "networking"
                ],
                "policy_vendor": "ubuntu-snappy",
                "policy_version": 88.88
            }
            """))
        with io.open(os.path.join(contents, 'meta', 'app2.apparmor'),
                  encoding='utf-8') as fp:
            data = fp.read()
        self.assertEqual(data, dedent("""\
            {
                "template": "default",
                "policy_groups": [
                    "networking"
                ],
                "policy_vendor": "ubuntu-snappy",
                "policy_version": 1.3
            }
            """))

    def test_one_explicit_binary_apparmor_file(self):
        self.use_temp_dir()
        scratch = os.path.join(self.temp_dir, "scratch")
        touch(os.path.join(scratch, "bin", "foo"))
        touch(os.path.join(scratch, "data", "icon.svg"))
        apparmor_path = os.path.join(scratch, 'path', 'to', 'cli1.apparmor')
        os.makedirs(os.path.dirname(apparmor_path))
        with io.open(apparmor_path, 'w', encoding='utf-8') as fp:
            fp.write(dedent(u"""\
            {
                "template": "default",
                "policy_groups": [
                    "networking"
                ],
                "policy_vendor": "ubuntu-snappy",
                "policy_version": 99.99
            }
            """))
        self._make_metadata(scratch, APPARMOR_3_TEMPLATE)
        self.builder.add_file(scratch, "./")
        # FIXME: make it a .snap
        expected_path = os.path.join(self.temp_dir, "test-yaml_1.0_all.click")
        actual_path = self.builder.build(self.temp_dir, "meta/package.yaml")
        self.assertEqual(expected_path, actual_path)
        self.assertTrue(os.path.exists(actual_path))
        contents = os.path.join(self.temp_dir, 'contents')
        subprocess.check_output(
            ["dpkg-deb", "-R", actual_path, contents], universal_newlines=True)
        self.assertEqual(sorted(os.listdir(contents)),
                         ['DEBIAN', 'bin', 'data', 'meta', 'path'])
        self.assertEqual(
            sorted(os.listdir(os.path.join(contents, 'path', 'to'))),
            ['cli1.apparmor'])
        with io.open(os.path.join(contents, 'path', 'to', 'cli1.apparmor'),
                     encoding='utf-8') as fp:
            data = fp.read()
        self.assertEqual(data, dedent("""\
            {
                "template": "default",
                "policy_groups": [
                    "networking"
                ],
                "policy_vendor": "ubuntu-snappy",
                "policy_version": 99.99
            }
            """))
        with io.open(os.path.join(contents, 'meta', 'cli2.apparmor'),
                     encoding='utf-8') as fp:
            data = fp.read()
        self.assertEqual(data, dedent("""\
            {
                "template": "default",
                "policy_groups": [
                    "networking"
                ],
                "policy_vendor": "ubuntu-snappy",
                "policy_version": 1.3
            }
            """))

    def test_one_explicit_binary_apparmor_profile_file(self):
        self.use_temp_dir()
        scratch = os.path.join(self.temp_dir, "scratch")
        touch(os.path.join(scratch, "bin", "foo"))
        touch(os.path.join(scratch, "data", "icon.svg"))
        apparmor_path = os.path.join(
            scratch, 'path', 'to', 'cli1.apparmor-profile')
        os.makedirs(os.path.dirname(apparmor_path))
        with io.open(apparmor_path, 'w', encoding='utf-8') as fp:
            fp.write(dedent(u"""\
            {
                "template": "default",
                "policy_groups": [
                    "networking"
                ],
                "policy_vendor": "ubuntu-snappy",
                "policy_version": 88.88
            }
            """))
        self._make_metadata(scratch, APPARMOR_4_TEMPLATE)
        self.builder.add_file(scratch, "./")
        # FIXME: make it a .snap
        expected_path = os.path.join(self.temp_dir, "test-yaml_1.0_all.click")
        actual_path = self.builder.build(self.temp_dir, "meta/package.yaml")
        self.assertEqual(expected_path, actual_path)
        self.assertTrue(os.path.exists(actual_path))
        contents = os.path.join(self.temp_dir, 'contents')
        subprocess.check_output(
            ["dpkg-deb", "-R", actual_path, contents], universal_newlines=True)
        self.assertEqual(sorted(os.listdir(contents)),
                         ['DEBIAN', 'bin', 'data', 'meta', 'path'])
        self.assertEqual(
            sorted(os.listdir(os.path.join(contents, 'path', 'to'))),
            ['cli1.apparmor-profile'])
        with io.open(
                os.path.join(contents, 'path', 'to', 'cli1.apparmor-profile'),
                encoding='utf-8') as fp:
            data = fp.read()
        self.assertEqual(data, dedent("""\
            {
                "template": "default",
                "policy_groups": [
                    "networking"
                ],
                "policy_vendor": "ubuntu-snappy",
                "policy_version": 88.88
            }
            """))
        with io.open(os.path.join(contents, 'meta', 'cli2.apparmor'),
                  encoding='utf-8') as fp:
            data = fp.read()
        self.assertEqual(data, dedent("""\
            {
                "template": "default",
                "policy_groups": [
                    "networking"
                ],
                "policy_vendor": "ubuntu-snappy",
                "policy_version": 1.3
            }
            """))

    def test_empty_readme(self):
        # LP: #1400308, case 1
        self.use_temp_dir()
        self._make_metadata(self.temp_dir, readme='')
        with self.assertRaises(ClickBuildError) as cm:
            self.builder.read_manifest(self.temp_dir + "/meta/package.yaml")
            self.assertEqual(
                str(cm.exception)[-59:],
                'meta/package.yaml must contain at least two non-empty lines')

    def test_readme_with_nonempty_blank_first_line(self):
        # LP: #1400308, case 1; first line is .strip()'d
        self.use_temp_dir()
        self._make_metadata(self.temp_dir, readme='          ')
        with self.assertRaises(ClickBuildError) as cm:
            self.builder.read_manifest(self.temp_dir + "/meta/package.yaml")
            self.assertEqual(
                str(cm.exception)[-59:],
                'meta/package.yaml must contain at least two non-empty lines')

    def test_single_line_readme(self):
        # LP: #1400308, case 2
        self.use_temp_dir()
        self._make_metadata(self.temp_dir, readme='a title\n')
        with self.assertRaises(ClickBuildError) as cm:
            self.builder.read_manifest(self.temp_dir + "/meta/package.yaml")
            self.assertEqual(
                str(cm.exception)[-59:],
                'meta/package.yaml must contain at least two non-empty lines')
